#include "stevedore.h"
#include <time.h>
#include <pthread.h>
#include "tallyman_mib.h"

#define MAX_SERVICE 1024

unsigned ttl_instance_state = 60;

struct service {
	char *id;
	struct instance *head, *tail;
	pthread_mutex_t mutex;
};

enum instance_state {
	instance_state_stopped,
	instance_state_running,
	instance_state_expired,
	instance_state_error
};

struct instance {
	char *hostname;
	struct service *srv;
	enum instance_state state;
	time_t timestamp;
	char *error_message;
	struct instance *prev, *next;
	pthread_cond_t notbusy;     /* Prevent simultaneous updates */
	int busy:1;                 /* Node is in use if 1 */
};

static struct service service[MAX_SERVICE];
static size_t service_count;

enum {
	SERVICE_TABLE_VALID = 0x01,
	INSTANCE_TABLE_VALID = 0x02
};

static int table_validity_state;
static pthread_mutex_t table_validity_mutex = PTHREAD_MUTEX_INITIALIZER;

static int
table_validity_start(int what)
{
	int result;
	pthread_mutex_lock(&table_validity_mutex);
	result = table_validity_state & what;
	table_validity_state |= what;
	return result;
}

static void
table_validity_end(void)
{
	pthread_mutex_unlock(&table_validity_mutex);
}

static void
table_validity_clear(void)
{
	pthread_mutex_lock(&table_validity_mutex);
	table_validity_state = 0;
	pthread_mutex_unlock(&table_validity_mutex);
}

struct service *
service_by_id(char const *id)
{
	size_t i;

	for (i = 0; i < service_count; i++)
		if (strcmp(service[i].id, id) == 0)
			return &service[i];
	return 0;
}

int
service_add(char const *id, grecs_locus_t *loc)
{
	if (service_by_id(id)) {
		grecs_error(loc, 0, "service %s already registered", id);
		return 1;
	}
	
	if (service_count == MAX_SERVICE) {
		grecs_error(loc, 0, "service table overflow");
		exit(1);
	}

	service[service_count].id = grecs_strdup(id);
	service[service_count].head = service[service_count].tail = NULL;
	pthread_mutex_init(&service[service_count].mutex, NULL);
	
	service_count++;

	return 0;
}

static struct instance *
instance_alloc(char const *hostname)
{
	struct instance *inst = calloc(1, sizeof(*inst));

	if (!inst) {
		grecs_error(NULL, 0, "out of memory");
		return NULL;
	}

	if ((inst->hostname = strdup(hostname)) == NULL) {
		free(inst);
		grecs_error(NULL, 0, "out of memory");
		return NULL;
	}

	inst->next = inst->prev = NULL;

	inst->busy = 1;
	pthread_cond_init(&inst->notbusy, NULL);
	
	return inst;
}

static void
instance_lock(struct instance *inst)
{
	assert(inst->srv != NULL);
	if (inst->busy) {
		pthread_cond_wait(&inst->notbusy, &inst->srv->mutex);
	}
	inst->busy = 1;
}

static void
instance_unlock(struct instance *inst)
{
	assert(inst->srv != NULL);
	if (inst->busy) {
		inst->busy = 0;
		pthread_cond_broadcast(&inst->notbusy);
	}
}

static void instance_unlink(struct instance *inst);

static void
instance_free(struct instance *inst)
{
	if (inst) {
		instance_unlink(inst);
		free(inst->hostname);
		free(inst);
	}
}

static void
instance_link(struct instance *inst, struct service *srv)
{
	inst->srv = srv;
	inst->prev = srv->tail;
	if (srv->tail)
		srv->tail->next = inst;
	else
		srv->head = inst;
	srv->tail = inst;
}

static void
instance_unlink(struct instance *inst)
{
	struct service *srv = inst->srv;
	struct instance *p;

	if (!srv)
		return;
	
	p = inst->prev;
	if (p)
		p->next = inst->next;
	else
		srv->head = inst->next;

	p = inst->next;
	if (p)
		p->prev = inst->prev;
	else
		srv->tail = inst->prev;

	inst->srv = NULL;
	inst->next = inst->prev = NULL;
}

static inline int
instance_is_running(struct instance *inst)
{
	return inst->state == instance_state_running;
}
	
static struct instance *
instance_by_hostname(char const *hostname, struct service *srv)
{
	struct instance *inst;
	for (inst = srv->head; inst; inst = inst->next)
		if (strcmp(inst->hostname, hostname) == 0)
			return inst;
	return NULL;
}

static struct json_value *
json_value_lookup_typed(struct json_value *obj, char const *name,
			enum json_value_type type)
{
	struct json_value *v = json_value_lookup(obj, name);
	return (v && v->type == type) ? v : NULL;
}

int
servdb_update(struct json_value *obj)
{
	struct instance *inst;
	struct service *srv;
	struct json_value *val;
	int status;
	char const *hostname;
	char const *id;
	
	val = json_value_lookup_typed(obj, "status", json_bool);
	if (!val)
		return -1;
	status = val->v.b;
	
	val = json_value_lookup_typed(obj, "id", json_string);
	if (!val)
		return -1;
	id = val->v.s;

	val = json_value_lookup_typed(obj, "hostname", json_string);
	if (!val)
		return -1;
	hostname = val->v.s;

	srv = service_by_id(id);
	if (!srv)
		return -1;

	pthread_mutex_lock(&srv->mutex);
	inst = instance_by_hostname(hostname, srv);
	if (!inst) {
		inst = instance_alloc(hostname);
		if (!inst) {
			pthread_mutex_unlock(&srv->mutex);
			return -1;
		}
		instance_link(inst, srv);
	} else
		instance_lock(inst);
	
	time(&inst->timestamp);

	val = json_value_lookup_typed(obj, "error", json_bool);
	if (val && val->v.b) {
		inst->state = instance_state_error;
		val = json_value_lookup_typed(obj, "message", json_string);
		inst->error_message = val ? strdup(val->v.s) : NULL;
	} else {
		inst->state = status
			       ? instance_state_running
			       : instance_state_stopped;
	}
	instance_unlock(inst);
	pthread_mutex_unlock(&srv->mutex);
	table_validity_clear();
	return 0;
}

size_t
servdb_servicesTotal(void)
{
	return service_count;
}

size_t
servdb_servicesRunning(void)
{
	size_t i, result = 0;
	for (i = 0; i < service_count; i++)
		if (servdb_serviceInstances(i))
			result++;
	return result;
}

ssize_t
servdb_serviceInstances(size_t idx)
{
	struct instance *inst;
	ssize_t result = 0;
	
	if (idx > service_count)
		return -1;
	pthread_mutex_lock(&service[idx].mutex);
	for (inst = service[idx].head; inst; inst = inst->next) {
		if (instance_is_running(inst)) {
			result++;
		}
	}
	pthread_mutex_unlock(&service[idx].mutex);
	return result;
}

static int
service_Table_load_row(size_t idx, netsnmp_tdata *table)
{
	netsnmp_tdata_row *data_row;
	struct serviceTable_entry *ent;
	struct service *srv = &service[idx];
		
	ent = SNMP_MALLOC_TYPEDEF(struct serviceTable_entry);
	if (!ent)
		return SNMP_ERR_GENERR;

	data_row = netsnmp_tdata_create_row();
	if (!data_row) {
		SNMP_FREE(ent);
		return SNMP_ERR_GENERR;
	}
	data_row->data = ent;

	ent->serviceIndex = idx;
	ent->serviceName = srv->id;
	ent->serviceName_len = strlen(srv->id);
	ent->serviceInstances = servdb_serviceInstances(idx);

	netsnmp_tdata_row_add_index(data_row, ASN_INTEGER,
				    &ent->serviceIndex,
				    sizeof(ent->serviceIndex));
	netsnmp_tdata_add_row(table, data_row);

	return SNMP_ERR_NOERROR;
}
	
int
serviceTable_load(netsnmp_cache *cache, void *vmagic)
{
	int rc;
	size_t i;
	netsnmp_tdata *table = (netsnmp_tdata *) vmagic;

	if (!table_validity_start(SERVICE_TABLE_VALID)) {
		serviceTable_free(cache, vmagic);
		for (i = 0; i < service_count; i++) {
			if ((rc = service_Table_load_row(i, table)) != SNMP_ERR_NOERROR)
				return rc;
		}
	}
	table_validity_end();
	return 0;
}

void
serviceTable_entry_free(void *data)
{
	/* nothing */
}

#define TMSEC(t) (((t)->tm_hour * 60 + (t)->tm_min) * 60 + (t)->tm_sec)

static int
utc_offset (void)
{
	time_t t = time (NULL);
	struct tm ltm = *localtime (&t);
	struct tm gtm = *gmtime (&t);
	int d = TMSEC (&ltm) - TMSEC (&gtm);
	if (!(ltm.tm_year = gtm.tm_year && ltm.tm_mon == gtm.tm_mon))
		d += 86400;
	return d / 60;
}

static int
setDateAndTime(time_t t, char **pbuf, size_t *plen)
{
	char *buf;
	size_t len = 11;
	struct tm *tm;
	unsigned n;
	
	buf = malloc(len + 1);
	if (!buf) 
		return -1;
	
	tm = localtime(&t);
	/*    A date-time specification.
          
	      field  octets  contents                  range
	      -----  ------  --------                  -----
	        1      1-2   year*                     0..65536
		2       3    month                     1..12
		3       4    day                       1..31
		4       5    hour                      0..23
		5       6    minutes                   0..59
		6       7    seconds                   0..60
		(use 60 for leap-second)
		7       8    deci-seconds              0..9
		8       9    direction from UTC        '+' / '-'
		9      10    hours from UTC*           0..13
		10      11    minutes from UTC          0..59
		
		* Notes:
		- the value of year is in network-byte order
	*/
	n = tm->tm_year % 100;
	buf[0] = n >> 8;
	buf[1] = n & 0xff;
	buf[2] = tm->tm_mon + 1;
	buf[3] = tm->tm_mday;
	buf[4] = tm->tm_hour; 
	buf[5] = tm->tm_min;
	buf[6] = tm->tm_sec;
	buf[7] = '0';
	n = utc_offset();
	if (n < 0) {
		buf[8] = '-';
		n = - n;
	} else 
		buf[8] = '+';
	buf[9] = n / 60;
	buf[10] = n % 60;

	*pbuf = buf;
	*plen = len;

	return 0;
}

static int
instanceTable_entry_load(size_t idx,
			 struct instance *inst, char const *service_id,
			 struct instanceTable_entry *ent)
{
	memset(ent, 0, sizeof(*ent));
	ent->instanceIndex = idx;
	ent->instanceName = strdup(inst->hostname);
	if (!ent->instanceName)
		return -1;
	ent->instanceName_len = strlen(ent->instanceName);
	ent->instanceService = (char*) service_id;
	ent->instanceService_len = strlen(service_id);
	ent->instanceState = inst->state;
	setDateAndTime(inst->timestamp, &ent->instanceTimeStamp,
		       &ent->instanceTimeStamp_len);
	ent->instanceErrorMessage = strdup(inst->error_message
					   ? inst->error_message
					   : "");
	if (!ent->instanceErrorMessage)
		return -1;
	ent->instanceErrorMessage_len = strlen(ent->instanceErrorMessage);
	return 0;
}
	
static int
instanceTable_load_row(size_t idx,
		       struct instance *inst, char const *service_id,
		       netsnmp_tdata *table)
{
	netsnmp_tdata_row *data_row;
	struct instanceTable_entry *ent;
	
	ent = SNMP_MALLOC_TYPEDEF(struct instanceTable_entry);
	if (!ent) 
		return SNMP_ERR_GENERR;

	if (instanceTable_entry_load(idx, inst, service_id, ent)) {
		instanceTable_entry_free(ent);
		SNMP_FREE(ent);
		return SNMP_ERR_GENERR;
	}
		
	data_row = netsnmp_tdata_create_row();
	if (!data_row) {
		instanceTable_entry_free(ent);
		SNMP_FREE(ent);
		return SNMP_ERR_GENERR;
	}
	data_row->data = ent;
	netsnmp_tdata_row_add_index(data_row, ASN_INTEGER,
				    &ent->instanceIndex,
				    sizeof(ent->instanceIndex));
	netsnmp_tdata_add_row(table, data_row);

	return SNMP_ERR_NOERROR;
}

int
instanceTable_load(netsnmp_cache *cache, void *vmagic)
{
	size_t i, idx;
	netsnmp_tdata *table = (netsnmp_tdata *) vmagic;
	int rc = SNMP_ERR_NOERROR;

	if (!table_validity_start(INSTANCE_TABLE_VALID)) {
		instanceTable_free(cache, vmagic);
		idx = 0;
		for (i = 0; i < service_count && rc == SNMP_ERR_NOERROR; i++) {
			struct instance *inst;
		
			pthread_mutex_lock(&service[i].mutex);
			for (inst = service[i].head; inst; inst = inst->next) {
				instance_lock(inst);
				rc = instanceTable_load_row(idx, inst,
							    service[i].id,
							    table);
				instance_unlock(inst);
				if (rc != SNMP_ERR_NOERROR)
					break;
				idx++;
			}
			pthread_mutex_unlock(&service[i].mutex);
		}
	}
	table_validity_end();
	return rc;
}

void
instanceTable_entry_free(void *data)
{
	struct instanceTable_entry *ent = data;
	free(ent->instanceName);
	free(ent->instanceErrorMessage);
}

void *
invalidator_thread(void *p)
{
	size_t i;
	struct instance *inst;
	while (1) {
		time_t now = time(NULL);
		int invalidate = 0;
		for (i = 0; i < service_count; i++) {
			pthread_mutex_lock(&service[i].mutex);
			for (inst = service[i].head; inst; inst = inst->next) {
				instance_lock(inst);
				if (inst->state != instance_state_expired
				    && now - inst->timestamp > ttl_instance_state)
					inst->state = instance_state_expired;
				instance_unlock(inst);
				invalidate = 1;
			}
			pthread_mutex_unlock(&service[i].mutex);
		}
		if (invalidate)
			table_validity_clear();
		sleep(ttl_instance_state);
	}
	return NULL;
}
